@@ -9,7 +9,7 @@ impl Config {
     pub async fn run(&self) -> Result<()> {
         warn!("this output is almost certain to change in a future version");
         let cfg = HKConfig::get()?;
-        println!("{}", cfg);
+        println!("{cfg}");
         Ok(())
     }
 }
