@@ -20,14 +20,14 @@ impl Init {
     pub async fn run(&self) -> Result<()> {
         let hk_file = PathBuf::from("hk.pkl");
         let hook_content = r#"
-amends "package://github.com/jdx/hk/releases/download/v0.8.5/hk@0.8.5#/Config.pkl"
-import "package://github.com/jdx/hk/releases/download/v0.8.5/hk@0.8.5#/Builtins.pkl"
+amends "package://github.com/jdx/hk/releases/download/v1.2.0/hk@1.2.0#/Config.pkl"
+import "package://github.com/jdx/hk/releases/download/v1.2.0/hk@1.2.0#/Builtins.pkl"
 
 local linters = new Mapping<String, Step> {
     // uses builtin prettier linter config
     ["prettier"] = Builtins.prettier
 
-    // uses custom pkl linter config
+    // define a custom linter
     ["pkl"] {
         glob = "*.pkl"
         check = "pkl eval {{files}} >/dev/null"
@@ -36,8 +36,8 @@ local linters = new Mapping<String, Step> {
 
 hooks {
     ["pre-commit"] {
-        fix = true           // automatically modify files with available linter fixes
-        stash = "patch-file" // stashes unstaged changes while running fix steps
+        fix = true    // automatically modify files with available linter fixes
+        stash = "git" // stashes unstaged changes while running fix steps
         steps {
             // "prelint" here is simply a name to define the step
             ["prelint"] {
@@ -68,7 +68,7 @@ hooks {
 "#;
         if !hk_file.exists() || self.force {
             xx::file::write(hk_file, hook_content.trim_start())?;
-        } else if !self.force {
+        } else if hk_file.exists() {
             warn!("hk.pkl already exists, run with --force to overwrite");
         }
 
