@@ -1,7 +1,7 @@
 use std::{
     cell::OnceCell,
     collections::BTreeSet,
-    ffi::OsString,
+    ffi::{CString, OsString},
     path::{Path, PathBuf},
     sync::Arc,
 };
@@ -9,13 +9,96 @@ use std::{
 use crate::Result;
 use clx::progress::{ProgressJob, ProgressJobBuilder, ProgressStatus};
 use eyre::{WrapErr, eyre};
-use git2::{Repository, StatusOptions, StatusShow, Tree};
+use git2::{Repository, StatusOptions, StatusShow};
 use itertools::Itertools;
 use serde::{Deserialize, Serialize};
+#[cfg(unix)]
+use std::os::unix::ffi::OsStringExt;
 use xx::file::display_path;
 
 use crate::env;
 
+fn parse_paths_from_patch(patch: &str) -> Vec<PathBuf> {
+    // Parse lines like: diff --git a/path b/path
+    let mut files = BTreeSet::new();
+    for line in patch.lines() {
+        if let Some(rest) = line.strip_prefix("diff --git a/") {
+            if let Some((a_path, _b)) = rest.split_once(" b/") {
+                let p = PathBuf::from(a_path);
+                if p.exists() {
+                    files.insert(p);
+                }
+            }
+        }
+    }
+    files.into_iter().collect()
+}
+
+fn conflicted_files_from_porcelain(status_z: &str) -> Vec<PathBuf> {
+    // In porcelain v1 short format: lines start with two status letters.
+    // Unmerged statuses include: UU, AA, AU, UA, DU, UD
+    let mut files = BTreeSet::new();
+    for entry in status_z.split('\0').filter(|s| !s.is_empty()) {
+        let mut chars = entry.chars();
+        let x = chars.next().unwrap_or(' ');
+        let y = chars.next().unwrap_or(' ');
+        let path: String = chars.skip(1).collect();
+        let is_unmerged = matches!(
+            (x, y),
+            ('U', 'U') | ('A', 'A') | ('A', 'U') | ('U', 'A') | ('D', 'U') | ('U', 'D')
+        );
+        if is_unmerged {
+            let p = PathBuf::from(path);
+            if p.exists() {
+                files.insert(p);
+            }
+        }
+    }
+    files.into_iter().collect()
+}
+
+fn resolve_conflict_markers_preferring_theirs(path: &Path) -> Result<()> {
+    let content = xx::file::read_to_string(path).unwrap_or_default();
+    if !content.contains("<<<<<<<") || !content.contains(">>>>>>>") {
+        return Ok(());
+    }
+
+    // Process as UTF-8 text, preserving original Unicode outside conflict blocks
+    // and keeping exact line endings by using split_inclusive.
+    let mut output = String::with_capacity(content.len());
+    let parts: Vec<&str> = content.split_inclusive('\n').collect();
+    let mut idx = 0usize;
+    while idx < parts.len() {
+        let line = parts[idx];
+        if line.starts_with("<<<<<<<") {
+            // Advance past our side until the separator '======='
+            idx += 1;
+            while idx < parts.len() && !parts[idx].starts_with("=======") {
+                idx += 1;
+            }
+            // Skip the separator line if present
+            if idx < parts.len() && parts[idx].starts_with("=======") {
+                idx += 1;
+            }
+            // Capture the 'theirs' side until the closing '>>>>>>>'
+            while idx < parts.len() && !parts[idx].starts_with(">>>>>>>") {
+                output.push_str(parts[idx]);
+                idx += 1;
+            }
+            // Skip the closing marker if present
+            if idx < parts.len() && parts[idx].starts_with(">>>>>>>") {
+                idx += 1;
+            }
+        } else {
+            output.push_str(line);
+            idx += 1;
+        }
+    }
+
+    xx::file::write(path, output)?;
+    Ok(())
+}
+
 pub struct Git {
     repo: Option<Repository>,
     stash: Option<StashType>,
@@ -46,6 +129,7 @@ impl Git {
             .ok_or(eyre!("failed to find git repository"))?;
         std::env::set_current_dir(&root)?;
         let repo = if *env::HK_LIBGIT2 {
+            debug!("libgit2: true");
             let repo = Repository::open(".").wrap_err("failed to open repository")?;
             if let Some(index_file) = &*env::GIT_INDEX_FILE {
                 // sets index to .git/index.lock which is used in the case of `git commit -a`
@@ -54,6 +138,7 @@ impl Git {
             }
             Some(repo)
         } else {
+            debug!("libgit2: false");
             None
         };
         Ok(Self {
@@ -116,58 +201,36 @@ impl Git {
         }
     }
 
-    fn head_tree(&self) -> Result<Tree<'_>> {
-        let head = self
-            .repo
-            .as_ref()
-            .unwrap()
-            .head()
-            .wrap_err("failed to get head")?;
-        let head = head
-            .peel_to_tree()
-            .wrap_err("failed to peel head to tree")?;
-        Ok(head)
-    }
-
-    // fn head_commit(&self) -> Result<Commit<'_>> {
-    //     let head = self.repo.head().wrap_err("failed to get head")?;
-    //     let commit = head
-    //         .peel_to_commit()
-    //         .wrap_err("failed to peel head to commit")?;
-    //     Ok(commit)
-    // }
-
-    // fn head_commit_message(&self) -> Result<String> {
-    //     let commit = self.head_commit()?;
-    //     let message = commit
-    //         .message()
-    //         .ok_or(eyre!("failed to get commit message"))?;
-    //     Ok(message.to_string())
-    // }
-
-    pub fn all_files(&self) -> Result<Vec<PathBuf>> {
-        // TODO: should this also show untracked files?
-        if let Some(_repo) = &self.repo {
-            let head = self.head_tree()?;
-            let mut files = Vec::new();
-            head.walk(git2::TreeWalkMode::PreOrder, |root, entry| {
-                if let Some(name) = entry.name() {
-                    let path = if root.is_empty() {
-                        PathBuf::from(name)
-                    } else {
-                        PathBuf::from(root).join(name)
-                    };
-                    if path.exists() {
-                        files.push(path);
+    pub fn all_files(&self, pathspec: Option<&[OsString]>) -> Result<BTreeSet<PathBuf>> {
+        // TODO: handle pathspec to improve globbing
+        if let Some(repo) = &self.repo {
+            let idx = repo.index()?;
+            Ok(idx
+                .iter()
+                .map(|i| {
+                    let cstr = CString::new(&i.path[..]).unwrap();
+                    #[cfg(unix)]
+                    {
+                        PathBuf::from(OsString::from_vec(cstr.as_bytes().to_vec()))
                     }
-                }
-                git2::TreeWalkResult::Ok
-            })
-            .wrap_err("failed to walk tree")?;
-            Ok(files)
+                    #[cfg(windows)]
+                    {
+                        PathBuf::from(cstr.into_string().unwrap())
+                    }
+                })
+                .collect())
         } else {
-            let output = xx::process::sh("git ls-files")?;
-            Ok(output.lines().map(PathBuf::from).collect())
+            let mut cmd = xx::process::cmd("git", ["ls-files", "-z"]);
+            if let Some(pathspec) = pathspec {
+                cmd = cmd.arg("--");
+                cmd = cmd.args(pathspec.iter().map(|p| p.to_str().unwrap()));
+            }
+            let output = cmd.read()?;
+            Ok(output
+                .split('\0')
+                .filter(|p| !p.is_empty())
+                .map(PathBuf::from)
+                .collect())
         }
     }
 
@@ -175,6 +238,7 @@ impl Git {
         if let Some(repo) = &self.repo {
             let mut status_options = StatusOptions::new();
             status_options.include_untracked(true);
+            status_options.recurse_untracked_dirs(true);
 
             if let Some(pathspec) = pathspec {
                 for path in pathspec {
@@ -223,21 +287,28 @@ impl Git {
                 modified_files,
             })
         } else {
-            let mut args = vec!["status", "--porcelain", "--no-renames"]
-                .into_iter()
-                .map(OsString::from)
-                .collect_vec();
+            let mut args = vec![
+                "status",
+                "--porcelain",
+                "--no-renames",
+                "--untracked-files=all",
+                "-z",
+            ]
+            .into_iter()
+            .filter(|&arg| !arg.is_empty())
+            .map(OsString::from)
+            .collect_vec();
             if let Some(pathspec) = pathspec {
                 args.push("--".into());
                 args.extend(pathspec.iter().map(|p| p.into()))
             }
-            let output = duct::cmd("git", args).read()?;
+            let output = xx::process::cmd("git", args).read()?;
             let mut staged_files = BTreeSet::new();
             let mut unstaged_files = BTreeSet::new();
             let mut untracked_files = BTreeSet::new();
             let mut modified_files = BTreeSet::new();
-            for line in output.lines() {
-                let mut chars = line.chars();
+            for file in output.split('\0') {
+                let mut chars = file.chars();
                 let index_status = chars.next().unwrap_or_default();
                 let workdir_status = chars.next().unwrap_or_default();
                 let path = PathBuf::from(chars.skip(1).collect::<String>());
@@ -344,11 +415,11 @@ impl Git {
                         .into_iter()
                         .chain(status.modified_files.iter().map(|p| p.to_str().unwrap()))
                         .collect::<Vec<_>>();
-                    duct::cmd("git", &args).run()?;
+                    xx::process::cmd("git", &args).run()?;
                 }
                 for file in status.untracked_files.iter() {
                     if let Err(err) = xx::file::remove_file(file) {
-                        warn!("failed to remove untracked file: {:?}", err);
+                        warn!("failed to remove untracked file: {err:?}");
                     }
                 }
             }
@@ -364,7 +435,9 @@ impl Git {
         let patch = if let Some(repo) = &self.repo {
             // essentially: git diff-index --ignore-submodules --binary --exit-code --no-color --no-ext-diff (git write-tree)
             let mut opts = git2::DiffOptions::new();
-            opts.include_untracked(true);
+            if *env::HK_STASH_UNTRACKED {
+                opts.include_untracked(true);
+            }
             opts.show_binary(true);
             opts.show_untracked_content(true);
             let diff = repo
@@ -388,27 +461,27 @@ impl Git {
                 String::from_utf8_lossy(&diff_bytes).to_string()
             }
         } else {
-            if !status.untracked_files.is_empty() {
+            if *env::HK_STASH_UNTRACKED && !status.untracked_files.is_empty() {
                 let args = vec!["add", "--intent-to-add", "--"]
                     .into_iter()
                     .chain(status.unstaged_files.iter().map(|p| p.to_str().unwrap()))
                     .collect::<Vec<_>>();
-                duct::cmd("git", &args).run()?;
+                xx::process::cmd("git", &args).run()?;
             }
             let output =
                 xx::process::sh("git diff --no-color --no-ext-diff --binary --ignore-submodules")?;
-            if !status.untracked_files.is_empty() {
+            if *env::HK_STASH_UNTRACKED && !status.untracked_files.is_empty() {
                 let args = vec!["reset", "--"]
                     .into_iter()
                     .chain(status.unstaged_files.iter().map(|p| p.to_str().unwrap()))
                     .collect::<Vec<_>>();
-                duct::cmd("git", &args).run()?;
+                xx::process::cmd("git", &args).run()?;
             }
             output
         };
         let patch_file = self.patch_file();
         if let Err(err) = xx::file::write(patch_file, &patch) {
-            warn!("failed to write patch file: {:?}", err);
+            warn!("failed to write patch file: {err:?}");
         }
         Ok(Some(StashType::PatchFile(patch, patch_file.to_path_buf())))
     }
@@ -420,13 +493,19 @@ impl Git {
         if let Some(repo) = &mut self.repo {
             let sig = repo.signature()?;
             let mut flags = git2::StashFlags::default();
-            flags.set(git2::StashFlags::INCLUDE_UNTRACKED, true);
+            if *env::HK_STASH_UNTRACKED {
+                flags.set(git2::StashFlags::INCLUDE_UNTRACKED, true);
+            }
             flags.set(git2::StashFlags::KEEP_INDEX, true);
             repo.stash_save(&sig, "hk", Some(flags))
                 .wrap_err("failed to stash")?;
             Ok(Some(StashType::LibGit))
         } else {
-            xx::process::sh("git stash push --keep-index --include-untracked -m hk")?;
+            let mut cmd = xx::process::cmd("git", ["stash", "push", "--keep-index", "-m", "hk"]);
+            if *env::HK_STASH_UNTRACKED {
+                cmd = cmd.arg("--include-untracked");
+            }
+            cmd.run()?;
             Ok(Some(StashType::Git))
         }
     }
@@ -448,18 +527,56 @@ impl Git {
                         ),
                     )
                     .start();
-                if let Some(repo) = &mut self.repo {
-                    let diff = git2::Diff::from_buffer(diff.as_bytes())?;
-                    let mut apply_opts = git2::ApplyOptions::new();
-                    repo.apply(&diff, git2::ApplyLocation::WorkDir, Some(&mut apply_opts))
-                        .wrap_err_with(|| {
-                            format!("failed to apply diff {}", display_path(&patch_file))
-                        })?;
-                } else {
-                    xx::process::sh(&format!("git apply {}", display_path(&patch_file)))?;
+                // Try a 3-way apply so non-conflicting hunks from the patch merge with
+                // any fixer changes. Conflict markers will be written for conflicting hunks.
+                let affected_files = parse_paths_from_patch(&diff);
+                let apply_res = xx::process::cmd("git", ["apply", "--3way"])
+                    .arg(&patch_file)
+                    .run();
+                // If 3-way apply reported an error, it might still have applied with conflicts.
+                // Detect conflict markers in affected files and treat that as a successful apply
+                // that requires resolution. Only fall back to plain apply when there are no
+                // conflict markers present.
+                let mut had_conflicts = false;
+                for f in &affected_files {
+                    if let Ok(s) = xx::file::read_to_string(f) {
+                        if s.contains("<<<<<<<") && s.contains(">>>>>>>") {
+                            had_conflicts = true;
+                            break;
+                        }
+                    }
+                }
+                if let Err(err) = apply_res {
+                    if had_conflicts {
+                        warn!(
+                            "git apply --3way returned error but left conflicts for {}: {err:?}; proceeding to resolve",
+                            display_path(&patch_file)
+                        );
+                    } else {
+                        warn!(
+                            "git apply --3way failed for {}: {err:?}; attempting plain apply",
+                            display_path(&patch_file)
+                        );
+                        if let Err(err2) = xx::process::cmd("git", ["apply"]).arg(&patch_file).run()
+                        {
+                            return Err(eyre!(
+                                "failed to apply patch (3-way and plain) {}: {err:?}; {err2:?}",
+                                display_path(&patch_file)
+                            ));
+                        }
+                    }
+                }
+                // Resolve any conflict markers by preferring the patch (stash) side
+                for f in affected_files {
+                    if let Err(err) = resolve_conflict_markers_preferring_theirs(&f) {
+                        warn!(
+                            "failed to resolve conflict markers in {}: {err:?}",
+                            display_path(&f)
+                        );
+                    }
                 }
                 if let Err(err) = xx::file::remove_file(patch_file) {
-                    debug!("failed to remove patch file: {:?}", err);
+                    debug!("failed to remove patch file: {err:?}");
                 }
             }
             // TODO: this does not work with untracked files
@@ -482,7 +599,60 @@ impl Git {
                 job = ProgressJobBuilder::new()
                     .prop("message", "stash – Applying git stash")
                     .start();
-                xx::process::sh("git stash pop")?;
+                // Apply the stash first, detect conflicts, and in case of conflict
+                // prefer the stash (the original unstaged changes), discarding fixer edits.
+                let apply_res = xx::process::cmd("git", ["stash", "apply"]).run();
+                if let Err(err) = &apply_res {
+                    warn!("git stash apply failed: {err:?}");
+                }
+                // Determine conflicted files via porcelain status
+                let status = match xx::process::cmd(
+                    "git",
+                    [
+                        "status",
+                        "--porcelain",
+                        "-z",
+                        "--no-renames",
+                        "--untracked-files=all",
+                    ],
+                )
+                .read()
+                {
+                    Ok(s) => s,
+                    Err(err) => {
+                        warn!("failed to read git status: {err:?}");
+                        String::new()
+                    }
+                };
+                let conflicted = conflicted_files_from_porcelain(&status);
+                if !conflicted.is_empty() {
+                    // For each conflicted file, prefer the stash side only for conflicting hunks
+                    for f in conflicted.iter() {
+                        if let Err(err) = resolve_conflict_markers_preferring_theirs(f) {
+                            warn!(
+                                "failed to resolve conflict markers in {}: {err:?}",
+                                display_path(f)
+                            );
+                        }
+                        if let Err(err) = xx::process::cmd("git", ["add", "--"]).arg(f).run() {
+                            warn!(
+                                "failed to stage {} after resolving conflicts: {err:?}",
+                                display_path(f)
+                            );
+                        }
+                    }
+                    // Drop the stash entry now that we've applied it
+                    if let Err(err) = xx::process::cmd("git", ["stash", "drop"]).run() {
+                        warn!("failed to drop stash after apply: {err:?}");
+                    }
+                } else if apply_res.is_err() {
+                    // Last resort: try pop
+                    if let Err(err) = xx::process::cmd("git", ["stash", "pop"]).run() {
+                        warn!("git stash pop also failed after apply error: {err:?}");
+                    }
+                } else if let Err(err) = xx::process::cmd("git", ["stash", "drop"]).run() {
+                    warn!("failed to drop stash after successful apply: {err:?}");
+                }
             }
         }
         job.set_status(ProgressStatus::Done);
@@ -500,12 +670,10 @@ impl Git {
             index.write().wrap_err("failed to write index")?;
             Ok(())
         } else {
-            let args = vec!["add", "--"]
-                .into_iter()
-                .map(OsString::from)
-                .chain(pathspecs.iter().map(|p| p.into()))
-                .collect::<Vec<_>>();
-            duct::cmd("git", &args).stdout_capture().run()?;
+            xx::process::cmd("git", ["add", "--"])
+                .args(pathspecs)
+                .stdout_capture()
+                .run()?;
             Ok(())
         }
     }
@@ -515,20 +683,28 @@ impl Git {
         if let Some(repo) = &self.repo {
             let from_obj = repo
                 .revparse_single(from_ref)
-                .wrap_err(format!("Failed to parse reference: {}", from_ref))?;
+                .wrap_err(format!("Failed to parse reference: {from_ref}"))?;
             let to_obj = repo
                 .revparse_single(to_ref)
-                .wrap_err(format!("Failed to parse reference: {}", to_ref))?;
+                .wrap_err(format!("Failed to parse reference: {to_ref}"))?;
 
-            let from_tree = from_obj
+            // Find the merge base between the two references
+            let merge_base = repo
+                .merge_base(from_obj.id(), to_obj.id())
+                .wrap_err("Failed to find merge base")?;
+            let merge_base_obj = repo
+                .find_object(merge_base, None)
+                .wrap_err("Failed to find merge base object")?;
+            let merge_base_tree = merge_base_obj
                 .peel_to_tree()
-                .wrap_err(format!("Failed to get tree for reference: {}", from_ref))?;
+                .wrap_err("Failed to get tree for merge base")?;
+
             let to_tree = to_obj
                 .peel_to_tree()
-                .wrap_err(format!("Failed to get tree for reference: {}", to_ref))?;
+                .wrap_err(format!("Failed to get tree for reference: {to_ref}"))?;
 
             let diff = repo
-                .diff_tree_to_tree(Some(&from_tree), Some(&to_tree), None)
+                .diff_tree_to_tree(Some(&merge_base_tree), Some(&to_tree), None)
                 .wrap_err("Failed to get diff between references")?;
 
             let mut files = BTreeSet::new();
@@ -550,11 +726,26 @@ impl Git {
 
             Ok(files.into_iter().collect())
         } else {
-            let output = xx::process::sh(&format!(
-                "git diff --name-only --diff-filter=ACRMT {}..{}",
-                from_ref, to_ref
-            ))?;
-            Ok(output.lines().map(PathBuf::from).collect())
+            // Use git merge-base to find the common ancestor
+            let merge_base = xx::process::sh(&format!("git merge-base {from_ref} {to_ref}"))?;
+            let merge_base = merge_base.trim();
+
+            let output = xx::process::cmd(
+                "git",
+                &[
+                    "diff",
+                    "-z",
+                    "--name-only",
+                    "--diff-filter=ACMRTUXB",
+                    format!("{merge_base}..{to_ref}").as_str(),
+                ],
+            )
+            .read()?;
+            Ok(output
+                .split('\0')
+                .filter(|p| !p.is_empty())
+                .map(PathBuf::from)
+                .collect())
         }
     }
 }
