@@ -1,5 +1,6 @@
-use crate::env;
+use crate::timings::StepTimingGuard;
 use crate::{Result, error::Error, step_job::StepJob};
+use crate::{env, step_job::StepJobStatus};
 use crate::{glob, settings::Settings};
 use crate::{step_context::StepContext, tera};
 use clx::progress::{ProgressJob, ProgressJobBuilder, ProgressJobDoneBehavior, ProgressStatus};
@@ -8,8 +9,10 @@ use eyre::{WrapErr, eyre};
 use indexmap::{IndexMap, IndexSet};
 use itertools::Itertools;
 use serde::{Deserialize, Serialize};
-use serde_with::serde_as;
-use std::{collections::HashSet, path::PathBuf};
+use serde_with::{DisplayFromStr, OneOrMany, PickFirst, serde_as};
+use shell_quote::QuoteInto;
+use shell_quote::QuoteRefExt;
+use std::{collections::HashSet, fmt::Display, path::PathBuf, str::FromStr};
 use std::{
     ffi::OsString,
     sync::{Arc, LazyLock},
@@ -18,10 +21,10 @@ use std::{fmt, process::Stdio};
 use tokio::sync::OwnedSemaphorePermit;
 use xx::file::display_path;
 
+#[serde_as]
 #[derive(Debug, Default, Clone, PartialEq, Eq, Deserialize, Serialize)]
 #[serde(rename_all = "snake_case")]
 #[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
-#[serde_as]
 pub struct Step {
     #[serde(default)]
     pub name: String,
@@ -32,10 +35,16 @@ pub struct Step {
     #[serde(default)]
     pub interactive: bool,
     pub depends: Vec<String>,
-    pub check: Option<String>,
-    pub check_list_files: Option<String>,
-    pub check_diff: Option<String>,
-    pub fix: Option<String>,
+    #[serde_as(as = "Option<PickFirst<(_, DisplayFromStr)>>")]
+    pub shell: Option<Script>,
+    #[serde_as(as = "Option<PickFirst<(_, DisplayFromStr)>>")]
+    pub check: Option<Script>,
+    #[serde_as(as = "Option<PickFirst<(_, DisplayFromStr)>>")]
+    pub check_list_files: Option<Script>,
+    #[serde_as(as = "Option<PickFirst<(_, DisplayFromStr)>>")]
+    pub check_diff: Option<Script>,
+    #[serde_as(as = "Option<PickFirst<(_, DisplayFromStr)>>")]
+    pub fix: Option<Script>,
     pub workspace_indicator: Option<String>,
     pub prefix: Option<String>,
     pub dir: Option<String>,
@@ -94,19 +103,19 @@ impl Step {
         }
     }
 
-    pub fn run_cmd(&self, run_type: RunType) -> Option<&str> {
+    pub fn run_cmd(&self, run_type: RunType) -> Option<&Script> {
         match run_type {
             RunType::Check(c) => match c {
-                CheckType::Check => self.check.as_deref(),
-                CheckType::Diff => self.check_diff.as_deref(),
-                CheckType::ListFiles => self.check_list_files.as_deref(),
+                CheckType::Check => self.check.as_ref(),
+                CheckType::Diff => self.check_diff.as_ref(),
+                CheckType::ListFiles => self.check_list_files.as_ref(),
             }
-            .or(self.check.as_deref())
-            .or(self.check_list_files.as_deref())
-            .or(self.check_diff.as_deref()),
+            .or(self.check.as_ref())
+            .or(self.check_list_files.as_ref())
+            .or(self.check_diff.as_ref()),
             RunType::Fix => self
                 .fix
-                .as_deref()
+                .as_ref()
                 .or_else(|| self.run_cmd(RunType::Check(CheckType::Check))),
         }
     }
@@ -181,9 +190,6 @@ impl Step {
         let mut workspaces: IndexSet<PathBuf> = Default::default();
         while let Some(dir) = dirs.pop() {
             if let Some(workspace) = xx::file::find_up(dir, &[workspace_indicator]) {
-                if let Some(dir) = dir.parent() {
-                    dirs.retain(|d| !d.starts_with(dir));
-                }
                 workspaces.insert(workspace);
             }
         }
@@ -197,18 +203,36 @@ impl Step {
             if files.is_empty() {
                 debug!("{self}: no files in {dir}");
             }
-            for f in files.iter_mut() {
-                // strip the dir prefix from the file path
-                *f = f.strip_prefix(dir).unwrap_or(f).to_path_buf();
-            }
+            // Don't strip the dir prefix here - it causes issues when steps have different working directories
+            // The path stripping should only happen in the command execution context via tera templates
         }
         if let Some(glob) = &self.glob {
-            files = glob::get_matches(glob, &files)?;
+            // when dir is set, globs are relative to that dir only
+            if let Some(dir) = &self.dir {
+                let dir_globs = glob
+                    .iter()
+                    .map(|g| format!("{}/{}", dir.trim_end_matches('/'), g))
+                    .collect::<Vec<_>>();
+                files = glob::get_matches(&dir_globs, &files)?;
+            } else {
+                files = glob::get_matches(glob, &files)?;
+            }
         }
         if let Some(exclude) = &self.exclude {
-            let excluded = glob::get_matches(exclude, &files)?
-                .into_iter()
-                .collect::<HashSet<_>>();
+            // when dir is set, excludes are relative to that dir only
+            let excluded = if let Some(dir) = &self.dir {
+                let dir_excludes = exclude
+                    .iter()
+                    .map(|g| format!("{}/{}", dir.trim_end_matches('/'), g))
+                    .collect::<Vec<_>>();
+                glob::get_matches(&dir_excludes, &files)?
+                    .into_iter()
+                    .collect::<HashSet<_>>()
+            } else {
+                glob::get_matches(exclude, &files)?
+                    .into_iter()
+                    .collect::<HashSet<_>>()
+            };
             files.retain(|f| !excluded.contains(f));
         }
         Ok(files)
@@ -227,11 +251,31 @@ impl Step {
             return Ok(Default::default());
         }
         let mut jobs = if let Some(workspace_indicators) = self.workspaces_for_files(&files)? {
-            let job = StepJob::new(Arc::new((*self).clone()), files.clone(), run_type);
+            let mut files = files.clone();
+
             workspace_indicators
-                .into_iter()
+                // Sort the files in reverse so the longest directory can take files in their directories
+                // and then the shortest path will take the rest of them.
+                .sorted_by(|a, b| b.as_os_str().len().cmp(&a.as_os_str().len()))
                 .map(|workspace_indicator| {
-                    job.clone().with_workspace_indicator(workspace_indicator)
+                    let workspace_dir = workspace_indicator.parent();
+                    let mut workspace_files = Vec::new();
+                    let mut i = 0;
+
+                    while i < files.len() {
+                        if workspace_dir
+                            .map(|dir| files[i].starts_with(dir))
+                            .unwrap_or(true)
+                        {
+                            let val = files.remove(i);
+                            workspace_files.push(val);
+                        } else {
+                            i += 1;
+                        }
+                    }
+
+                    StepJob::new(Arc::new((*self).clone()), workspace_files, run_type)
+                        .with_workspace_indicator(workspace_indicator)
                 })
                 .collect()
         } else if self.batch {
@@ -350,16 +394,24 @@ impl Step {
             return Ok(());
         }
         if matches!(ctx.hook_ctx.run_type, RunType::Fix) {
-            let stage = &self
+            // Build stage pathspecs; if `dir` is set, stage entries are relative to it
+            let stage: Vec<OsString> = self
                 .stage
                 .as_ref()
                 .unwrap_or(&vec![])
                 .iter()
                 .map(|s| tera::render(s, &ctx.hook_ctx.tctx).unwrap())
+                .map(|s| {
+                    if let Some(dir) = &self.dir {
+                        format!("{}/{}", dir.trim_end_matches('/'), s)
+                    } else {
+                        s
+                    }
+                })
                 .map(OsString::from)
                 .collect_vec();
             if !stage.is_empty() {
-                let status = ctx.hook_ctx.git.lock().await.status(Some(stage))?;
+                let status = ctx.hook_ctx.git.lock().await.status(Some(&stage))?;
                 let files = status.unstaged_files.into_iter().collect_vec();
                 if !files.is_empty() {
                     ctx.hook_ctx.git.lock().await.add(&files)?;
@@ -403,17 +455,15 @@ impl Step {
             }
         }
         job.progress = Some(job.build_progress(ctx));
-        if job.status.is_pending() {
-            let semaphore = if let Some(semaphore) = job.semaphore.take() {
-                semaphore
-            } else {
-                ctx.hook_ctx.semaphore().await
-            };
-            job.status_start(ctx, semaphore).await?;
-        }
+        job.status = StepJobStatus::Pending;
+        let semaphore = if let Some(semaphore) = job.semaphore.take() {
+            semaphore
+        } else {
+            ctx.hook_ctx.semaphore().await
+        };
+        job.status_start(ctx, semaphore).await?;
         let mut tctx = job.tctx(&ctx.hook_ctx.tctx);
         tctx.with_globs(self.glob.as_ref().unwrap_or(&vec![]));
-        tctx.with_files(&job.files);
         let file_msg = |files: &[PathBuf]| {
             format!(
                 "{} file{}",
@@ -422,10 +472,10 @@ impl Step {
             )
         };
         let Some(mut run) = self.run_cmd(job.run_type).map(|s| s.to_string()) else {
-            eyre::bail!("{}: no run command", self);
+            eyre::bail!("{self}: no run command");
         };
         if let Some(prefix) = &self.prefix {
-            run = format!("{} {}", prefix, run);
+            run = format!("{prefix} {run}");
         }
         let run = tera::render(&run, &tctx).unwrap();
         job.progress.as_ref().unwrap().prop(
@@ -443,10 +493,18 @@ impl Step {
                 trace!("{self}: {}", file.display());
             }
         }
-        let mut cmd = CmdLineRunner::new("sh")
-            .arg("-o")
-            .arg("errexit")
-            .arg("-c")
+        let mut cmd = if let Some(shell) = &self.shell {
+            let shell = shell.to_string();
+            let shell = shell.split_whitespace().collect_vec();
+            let mut cmd = CmdLineRunner::new(shell[0]);
+            for arg in shell[1..].iter() {
+                cmd = cmd.arg(arg);
+            }
+            cmd
+        } else {
+            CmdLineRunner::new("sh").arg("-o").arg("errexit").arg("-c")
+        };
+        cmd = cmd
             .arg(&run)
             .with_pr(job.progress.as_ref().unwrap().clone())
             .with_cancel_token(ctx.hook_ctx.failed.clone())
@@ -465,12 +523,21 @@ impl Step {
             let value = tera::render(value, &tctx)?;
             cmd = cmd.env(key, value);
         }
-        match cmd.execute().await {
+        let mut timing_guard = ctx
+            .hook_ctx
+            .timing
+            .clone()
+            .map(|t| StepTimingGuard::new(t, self));
+        let exec_result = cmd.execute().await;
+        if let Some(g) = timing_guard.take() {
+            g.finish();
+        }
+        if self.interactive {
+            clx::progress::resume();
+        }
+        match exec_result {
             Ok(_) => {}
             Err(err) => {
-                if self.interactive {
-                    clx::progress::resume();
-                }
                 if let ensembler::Error::ScriptFailed(e) = &err {
                     if let RunType::Check(CheckType::ListFiles) = job.run_type {
                         let result = &e.3;
@@ -489,15 +556,53 @@ impl Step {
                 return Err(err).wrap_err(run);
             }
         }
-        if self.interactive {
-            clx::progress::resume();
-        }
         ctx.decrement_job_count();
         job.status_finished()?;
         Ok(())
     }
+
+    pub fn shell_type(&self) -> ShellType {
+        let shell = self
+            .shell
+            .as_ref()
+            .map(|s| s.to_string())
+            .unwrap_or_default();
+        let shell = shell.split_whitespace().next().unwrap_or_default();
+        let shell = shell.split("/").last().unwrap_or_default();
+        match shell {
+            "bash" => ShellType::Bash,
+            "dash" => ShellType::Dash,
+            "fish" => ShellType::Fish,
+            "sh" => ShellType::Sh,
+            "zsh" => ShellType::Zsh,
+            _ => ShellType::Other(shell.to_string()),
+        }
+    }
 }
 
+pub enum ShellType {
+    Bash,
+    Dash,
+    Fish,
+    Sh,
+    Zsh,
+    #[allow(unused)]
+    Other(String),
+}
+
+impl ShellType {
+    pub fn quote(&self, s: &str) -> String {
+        match self {
+            ShellType::Bash | ShellType::Zsh => s.quoted(shell_quote::Bash),
+            ShellType::Fish => s.quoted(shell_quote::Fish),
+            ShellType::Dash | ShellType::Sh | ShellType::Other(_) => {
+                let mut o = vec![];
+                shell_quote::Sh::quote_into(s, &mut o);
+                String::from_utf8(o).unwrap_or_default()
+            }
+        }
+    }
+}
 fn is_profile_enabled(
     name: &str,
     enabled: Option<IndexSet<String>>,
@@ -559,3 +664,40 @@ fn try_canonicalize(path: &PathBuf) -> PathBuf {
         }
     }
 }
+
+#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
+#[serde_as]
+pub struct Script {
+    pub linux: Option<String>,
+    pub macos: Option<String>,
+    pub windows: Option<String>,
+    pub other: Option<String>,
+}
+
+impl FromStr for Script {
+    type Err = eyre::Error;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        Ok(Self {
+            linux: None,
+            macos: None,
+            windows: None,
+            other: Some(s.to_string()),
+        })
+    }
+}
+
+impl Display for Script {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let other = self.other.as_deref().unwrap_or_default();
+        if cfg!(target_os = "macos") {
+            write!(f, "{}", self.macos.as_deref().unwrap_or(other))
+        } else if cfg!(target_os = "linux") {
+            write!(f, "{}", self.linux.as_deref().unwrap_or(other))
+        } else if cfg!(target_os = "windows") {
+            write!(f, "{}", self.windows.as_deref().unwrap_or(other))
+        } else {
+            write!(f, "{other}")
+        }
+    }
+}
